use std::io::{stdin, stdout, BufRead, Write};
use std::iter::repeat;
use std::sync::mpsc::{self, Sender};
use std::thread;
use std::time::Duration;
use std::{collections::VecDeque, sync::mpsc::Receiver};

use rand::{seq::IteratorRandom, thread_rng};
use termion::terminal_size;
use termion::{clear, cursor::Goto};

fn main() {
	let mut screen = Screen::default();

	let lines = reading_thread();
	let mut queue = None;

	let mut stdout = stdout().lock();
	loop {
		// TODO : check for size
		let size = get_term_size();
		if screen.dim != size {
			screen.resize(size.0, size.1);
		}

		if queue.is_none() {
			if let Ok(received) = lines.try_recv() {
				queue = Some(received);
			}
		}

		if let Some((text, ok)) = &queue {
			let success = screen.append(text.clone());
			if success {
				ok.send(()).unwrap();
				queue.take();
			}
		}

		screen.step();
		stdout.write_all(screen.draw().as_bytes()).unwrap();
		stdout.flush().unwrap();
		thread::sleep(Duration::from_millis(100));
	}
}

fn reading_thread() -> Receiver<(String, Sender<()>)> {
	let (cmd_tx, cmd_rx) = mpsc::channel();
	thread::spawn(move || {
		let mut stdin = stdin().lock();
		loop {
			let mut line = String::new();
			stdin.read_line(&mut line).unwrap();
			let (tx, rx) = mpsc::channel();
			cmd_tx.send((line, tx)).unwrap();
			rx.recv().unwrap();
		}
	});
	cmd_rx
}

#[derive(Debug, Default)]
struct Screen {
	dim: (usize, usize),
	lines: Vec<ScreenLine>,
}

impl Screen {
	pub fn resize(&mut self, width: usize, height: usize) {
		let (old_width, _) = self.dim;
		let width_difference = width.saturating_sub(old_width);
		let new_lines = repeat(ScreenLine::default()).take(width_difference);
		self.lines.truncate(width);
		self.lines.extend(new_lines);
		self.dim = (width, height);
	}

	pub fn append(&mut self, text: String) -> bool {
		if let Some(available) = self.find_available_line() {
			available.append(text);
			true
		} else {
			false
		}
	}

	pub fn draw(&self) -> String {
		let clear = clear::All;
		let mut sequence = format!("{clear}");
		let (_, height) = self.dim;
		for (x, line) in self.lines.iter().enumerate() {
			for text in &line.content {
				for (pos, letter) in text.content.chars().enumerate() {
					let y = pos as isize + text.progress;
					if y < 0 || y >= (height - 1) as isize {
						continue;
					}
					let goto = Goto((x + 1) as u16, (y + 1) as u16);
					sequence += &format!("{goto}{letter}");
				}
			}
		}
		let goto_topleft = Goto(1, 1);
		sequence += &format!("{goto_topleft}");
		sequence
	}

	pub fn step(&mut self) {
		let (_, height) = self.dim;
		for line in &mut self.lines.iter_mut() {
			line.step(height);
		}
	}

	fn find_available_line(&mut self) -> Option<&mut ScreenLine> {
		self.lines
			.iter_mut()
			.filter(|l| l.is_available())
			.choose(&mut thread_rng())
	}
}

#[derive(Debug, Clone, Default)]
struct ScreenLine {
	content: VecDeque<Text>,
}

impl ScreenLine {
	pub fn is_available(&self) -> bool {
		let highest = self.content.back();
		highest.map(|t| t.progress >= 1).unwrap_or(true)
	}

	pub fn append(&mut self, text: String) {
		let content = text;
		let progress = -(content.len() as isize);
		self.content.push_back(Text { content, progress });
	}

	pub fn step(&mut self, height: usize) {
		for text in &mut self.content {
			text.progress += 1;
		}

		while let Some(last) = self.content.front() {
			if last.progress > height as isize {
				self.content.pop_front();
			} else {
				break;
			}
		}
	}
}

#[derive(Debug, Clone)]
struct Text {
	pub content: String,
	pub progress: isize,
}

fn get_term_size() -> (usize, usize) {
	let (width, height) = terminal_size().unwrap_or((80, 40));
	(width as usize, height as usize)
}
